"""
Copyright (C) 2017 Roberto Bruttomesso <roberto.bruttomesso@gmail.com>

This file is distributed under the terms of the 3-clause BSD License.
A copy of the license can be found in the root directory or at
https://opensource.org/licenses/BSD-3-Clause.

Author: Roberto Bruttomesso <roberto.bruttomesso@gmail.com>
  Date: 27/03/2017

This module implements the class for traces
"""

import pandas as pd
import intrepyd as ip

class Trace(object):
    """
    A trace generated from an Engine, or manually built for use in simulation
    """
    def __init__(self, ctx, rawtrace=None):
        self.ctx = ctx
        self.rawtrace = None
        if rawtrace is None:
            rawtrace = ip.api.mk_trace(self.ctx)
        self.rawtrace = rawtrace

    @staticmethod
    def get_numeric_value(value):
        """
        Coverts a string value (including true, false) into a corresponding float or integer.
        """
        if '.' in value:
            return float(value)
        elif value == '?':
            return '?'
        elif value == 'T':
            return 1
        elif value == 'F':
            return 0
        return int(value)

    def get_max_depth(self):
        """
        Returns the max depth of the trace
        """
        return ip.api.trace_get_max_depth(self.rawtrace)

    def get_value(self, net, depth):
        """
        Retrieval of a value from the trace for a net at a depth
        """
        size = ip.api.trace_prepare_value_for_net(self.ctx, self.rawtrace, net, depth)
        value = ''
        for i in range(size):
            value += ip.api.value_at(i)
        return value

    def set_value(self, net, depth, value):
        """
        Sets a value for a net at a certain depth
        """
        if value == 'true':
            value = 'T'
        elif value == 'false':
            value = 'F'
        ip.api.trace_set_value(self.ctx, self.rawtrace, net, depth, value)

    def get_as_net_dictionary(self, net2name=None):
        """
        Return the trace as a dictionary

        net -> [value@0, value@1, ...]
        """
        watched_nets = [ip.api.trace_get_watched_net(self.rawtrace, i)\
                       for i in range(ip.api.trace_get_watched_nets_number(self.rawtrace))]
        if net2name is None:
            return {net : list(self._get_as_steps(net)) for net in watched_nets}
        else:
            return {net2name[net] : list(self._get_as_steps(net)) for net in watched_nets}

    def get_as_depth_dictionary(self):
        """
        Return the trace as a dictionary

        depth -> [valuenet1@depth, valuenet2@depth, ...]
        """
        return {depth : list(self._get_values_at_depth(depth))\
                for depth in range(ip.api.trace_get_max_depth(self.rawtrace))}

    def get_as_dataframe(self, net2name):
        """
        Return the trace as a pandas dataframe

                  0       1       2    ...
        name1  value@0 value@1 value@2 ...
        name2  value@0 value@1 value@2 ...
        ...
        """
        matrix = []
        indexes = []
        for net, values in self.get_as_net_dictionary().iteritems():
            indexes.append(net2name[net])
            matrix.append(values)
        return pd.DataFrame(matrix, index=indexes)

    def set_from_dataframe(self, dataframe, inputname2net):
        """
        Return a trace from a pandas dataframe. Only input values
        are considered in filling up the trace
        """
        max_depth = 0
        row = 0
        for name in dataframe.index:
            if name in inputname2net:
                net = inputname2net[name]
                depth = 0
                for value in dataframe.values[row]:
                    self.set_value(net, depth, str(value))
                    if depth > max_depth:
                        max_depth = depth
                    depth += 1
            row += 1
        return max_depth

    def _get_as_steps(self, net):
        """
        Simplifies the retrieval of the set values per step for a net in a trace.
        """
        max_depth = ip.api.trace_get_max_depth(self.rawtrace)
        return (self.get_value(net, depth) for depth in range(0, max_depth + 1))

    def _get_values_at_depth(self, depth):
        """
        Simplifies the retrieval of the set values per step for a net in a trace.
        """
        watched_nets = [ip.api.trace_get_watched_net(self.rawtrace, i)\
                       for i in range(ip.api.trace_get_watched_nets_number(self.rawtrace))]
        return (self.get_value(net, depth) for net in watched_nets)



"""
Copyright (C) 2017 Roberto Bruttomesso <roberto.bruttomesso@gmail.com>

This file is distributed under the terms of the 3-clause BSD License.
A copy of the license can be found in the root directory or at
https://opensource.org/licenses/BSD-3-Clause.

Author: Roberto Bruttomesso <roberto.bruttomesso@gmail.com>
  Date: 27/03/2017

This module implements infrastructure to deal with SCR requirements
"""

import intrepyd as ip
import intrepyd.api
import csv
import os.path

def mk_raising_edge(ctx, edgeNet, pastEdgeNet, pastWhen):
    n1 = ctx.mk_not(pastEdgeNet)
    n2 = ctx.mk_and(n1, pastWhen)
    n3 = ctx.mk_and(n2, edgeNet)
    return n3

def mk_falling_edge(ctx, edgeNet, pastEdgeNet, pastWhen):
    n1 = ctx.mk_not(edgeNet)
    n2 = ctx.mk_and(n1, pastWhen)
    n3 = ctx.mk_and(n2, pastEdgeNet)
    return n3

def mk_row_condition(ctx, inputs, pastInputs, modeStr2mode, modes, pastMode, row, rowNumber):
    assert len(inputs) == len(pastInputs)
    assert len(row) == len(inputs) + 2

    oldModeIdx = modeStr2mode[row[0]]

    if oldModeIdx == None:
        raise Exception('Cannot find mode:', row[0])

    oldModeNet = modes[oldModeIdx]

    raisingEdge = False
    fallingEdge = False
    edgeNet = None
    pastEdgeNet = None
    pastWhen = ctx.mk_eq(oldModeNet, pastMode) 

    for i in range(1, len(row) - 1):
        conj = None
        if row[i] == 't':
            conj = pastInputs[i-1]
        elif row[i] == 'f':
            conj = ctx.mk_not(pastInputs[i-1])
        elif row[i] == 'T':
            raisingEdge = True
            edgeNet = inputs[i-1]
            pastEdgeNet = pastInputs[i-1]
        elif row[i] == 'F':
            fallingEdge = True
            edgeNet = inputs[i-1]
            pastEdgeNet = pastInputs[i-1]
        else:
            pass
        if conj != None:
            pastWhen = ctx.mk_and(pastWhen, conj)

    if edgeNet == None or pastEdgeNet == None:
        raise Exception('Cannot find raising or falling edge in row ' + str(rowNumber))

    if raisingEdge:
        return mk_raising_edge(ctx, edgeNet, pastEdgeNet, pastWhen)

    if fallingEdge:
        return mk_falling_edge(ctx, edgeNet, pastEdgeNet, pastWhen)

    raise Exception('Unreachable location (is apparently reachable)')

def retrieve_modes_order(ordfilename):
    modeStr2mode = {}
    modeStr2modeValue = {}
    lines = []
    with open(ordfilename, 'r') as ordfile:
        lines = ordfile.readlines()
    i = 0
    for line in lines:
        pair = line.split()
        if len(pair) != 2:
            raise Exception('Parse error at line ' + str(i+1) + ' of ' + ordfilename)
        modeStr2mode[pair[0]] = i
        modeStr2modeValue[pair[0]] = pair[1]
        i += 1
    return modeStr2mode, modeStr2modeValue

def mk_scr_helper(ctx, csvfilename, modeStr2mode, modeStr2modeValue, inputs, pastInputs, modes, pastMode):
    firstRow = True
    result = pastMode
    rowNumber = 1
    with open(csvfilename, 'r') as csvfile:
        for row in csv.reader(csvfile, delimiter=','):
            if len(inputs) + 2 != len(row):
                raise Exception('Input number mismatch: expected ' + str(len(inputs)) + ', actual ' + str(len(row) - 2))
            if firstRow:
                firstRow = False
                continue
            rowNumber += 1
            rowCondition = mk_row_condition(ctx, inputs, pastInputs, modeStr2mode, modes, pastMode, row, rowNumber)
            rowCurrentModeStr = str(modeStr2modeValue[row[-1]])
            rowCurrentModeNet = ctx.mk_number(rowCurrentModeStr, ctx.mk_int8_type())
            if rowCurrentModeNet == None:
                raise Exception('Cannot find mode:', row[-1])
            result = ctx.mk_ite(rowCondition, rowCurrentModeNet, result)
    return result

def mk_scr(ctx, name, inputs, pastInputs, modes, pastMode):
    if len(inputs) != len(pastInputs):
        raise Exception('Inputs and past inputs differ in size')
    csvfilename = name + '.csv'
    ordfilename = name + '.ord'
    if not os.path.isfile(csvfilename):
        raise Exception('Cannot find ' + csvfilename)
    if not os.path.isfile(ordfilename):
        raise Exception('Cannot find ' + ordfilename)
    modeStr2mode, modeStr2modeValue = retrieve_modes_order(ordfilename)
    return mk_scr_helper(ctx, csvfilename, modeStr2mode, modeStr2modeValue, inputs, pastInputs, modes, pastMode)
            





"""
Copyright (C) 2017 Roberto Bruttomesso <roberto.bruttomesso@gmail.com>

This file is distributed under the terms of the 3-clause BSD License.
A copy of the license can be found in the root directory or at
https://opensource.org/licenses/BSD-3-Clause.

Author: Roberto Bruttomesso <roberto.bruttomesso@gmail.com>
  Date: 22/05/2017

Retrieve values from the configuration file
"""
import json

class Config(object):
    """
    Stores configuration options
    """
    instance = None

    @staticmethod
    def get_instance(filename=None):
        """
        Implementation of Singleton pattern
        """
        if Config.instance is None:
            assert not filename is None
            Config.instance = Config(filename)
        return Config.instance.config

    def __init__(self, filename):
        with open(filename) as fdesc:
            config_string = ''
            for line in fdesc.readlines():
                line = line.lstrip()
                if len(line) > 2 and line[0] == '/' and line[1] == '/':
                    # Ignores commented-out lines
                    continue
                config_string += line
            self._config = json.loads(config_string)

    @property
    def config(self):
        """
        Getter
        """
        return self._config



"""
Copyright (C) 2017 Roberto Bruttomesso <roberto.bruttomesso@gmail.com>

This file is distributed under the terms of the 3-clause BSD License.
A copy of the license can be found in the root directory or at
https://opensource.org/licenses/BSD-3-Clause.

Author: Roberto Bruttomesso <roberto.bruttomesso@gmail.com>
  Date: 27/03/2017

This module implements the main parsing routine of Lustre files
"""

from antlr4 import FileStream, CommonTokenStream
from intrepyd.lustre2py.LustreLexer import LustreLexer
from intrepyd.lustre2py.LustreParser import LustreParser
from intrepyd.lustre2py.astbuilder import ASTBuilder
from intrepyd.lustre2py.ast import Ast

def parse(infile):
    """
    Parse the provided input file and translates
    into a python equivalent file for intrepyd
    """
    lexer = LustreLexer(FileStream(infile))
    stream = CommonTokenStream(lexer)
    parser = LustreParser(stream)
    tree = parser.program()
    ast_builder = ASTBuilder()
    ast_builder.visit(tree)
    return Ast(ast_builder.nodes)



# This file was automatically generated by SWIG (http://www.swig.org).
# Version 3.0.8
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.





from sys import version_info
if version_info >= (2, 6, 0):
    def swig_import_helper():
        from os.path import dirname
        import imp
        fp = None
        try:
            fp, pathname, description = imp.find_module('_api', [dirname(__file__)])
        except ImportError:
            import _api
            return _api
        if fp is not None:
            try:
                _mod = imp.load_module('_api', fp, pathname, description)
            finally:
                fp.close()
            return _mod
    _api = swig_import_helper()
    del swig_import_helper
else:
    import _api
del version_info
try:
    _swig_property = property
except NameError:
    pass  # Python < 2.2 doesn't have 'property'.


def _swig_setattr_nondynamic(self, class_type, name, value, static=1):
    if (name == "thisown"):
        return self.this.own(value)
    if (name == "this"):
        if type(value).__name__ == 'SwigPyObject':
            self.__dict__[name] = value
            return
    method = class_type.__swig_setmethods__.get(name, None)
    if method:
        return method(self, value)
    if (not static):
        if _newclass:
            object.__setattr__(self, name, value)
        else:
            self.__dict__[name] = value
    else:
        raise AttributeError("You cannot add attributes to %s" % self)


def _swig_setattr(self, class_type, name, value):
    return _swig_setattr_nondynamic(self, class_type, name, value, 0)


def _swig_getattr_nondynamic(self, class_type, name, static=1):
    if (name == "thisown"):
        return self.this.own()
    method = class_type.__swig_getmethods__.get(name, None)
    if method:
        return method(self)
    if (not static):
        return object.__getattr__(self, name)
    else:
        raise AttributeError(name)

def _swig_getattr(self, class_type, name):
    return _swig_getattr_nondynamic(self, class_type, name, 0)


def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)

try:
    _object = object
    _newclass = 1
except AttributeError:
    class _object:
        pass
    _newclass = 0



_api.INT_ENGINE_RESULT_UNKNOWN_swigconstant(_api)
INT_ENGINE_RESULT_UNKNOWN = _api.INT_ENGINE_RESULT_UNKNOWN

_api.INT_ENGINE_RESULT_REACHABLE_swigconstant(_api)
INT_ENGINE_RESULT_REACHABLE = _api.INT_ENGINE_RESULT_REACHABLE

_api.INT_ENGINE_RESULT_UNREACHABLE_swigconstant(_api)
INT_ENGINE_RESULT_UNREACHABLE = _api.INT_ENGINE_RESULT_UNREACHABLE

def mk_ctx():
    """mk_ctx() -> Int_ctx"""
    return _api.mk_ctx()

def del_ctx(ctx):
    """
    del_ctx(Int_ctx ctx)

    Parameters
    ----------
    ctx: Int_ctx

    """
    return _api.del_ctx(ctx)

def throw_exception(msg):
    """
    throw_exception(char const * msg)

    Parameters
    ----------
    msg: char const *

    """
    return _api.throw_exception(msg)

def clear_exception():
    """clear_exception()"""
    return _api.clear_exception()

def check_exception():
    """check_exception() -> char *"""
    return _api.check_exception()

def push_namespace(ctx, name):
    """
    push_namespace(Int_ctx ctx, char const * name)

    Parameters
    ----------
    ctx: Int_ctx
    name: char const *

    """
    return _api.push_namespace(ctx, name)

def pop_namespace(ctx):
    """
    pop_namespace(Int_ctx ctx)

    Parameters
    ----------
    ctx: Int_ctx

    """
    return _api.pop_namespace(ctx)

def mk_engine_bmc(ctx):
    """
    mk_engine_bmc(Int_ctx ctx) -> Int_engine_bmc

    Parameters
    ----------
    ctx: Int_ctx

    """
    return _api.mk_engine_bmc(ctx)

def set_bmc_current_depth(engine, depth):
    """
    set_bmc_current_depth(Int_engine_bmc engine, unsigned int depth)

    Parameters
    ----------
    engine: Int_engine_bmc
    depth: unsigned int

    """
    return _api.set_bmc_current_depth(engine, depth)

def set_bmc_optimize(engine):
    """
    set_bmc_optimize(Int_engine_bmc engine)

    Parameters
    ----------
    engine: Int_engine_bmc

    """
    return _api.set_bmc_optimize(engine)

def bmc_add_target(ctx, engine, target):
    """
    bmc_add_target(Int_ctx ctx, Int_engine_bmc engine, Int_net target)

    Parameters
    ----------
    ctx: Int_ctx
    engine: Int_engine_bmc
    target: Int_net

    """
    return _api.bmc_add_target(ctx, engine, target)

def bmc_add_watch(ctx, engine, watch):
    """
    bmc_add_watch(Int_ctx ctx, Int_engine_bmc engine, Int_net watch)

    Parameters
    ----------
    ctx: Int_ctx
    engine: Int_engine_bmc
    watch: Int_net

    """
    return _api.bmc_add_watch(ctx, engine, watch)

def bmc_reach_targets(engine):
    """
    bmc_reach_targets(Int_engine_bmc engine) -> Int_engine_result

    Parameters
    ----------
    engine: Int_engine_bmc

    """
    return _api.bmc_reach_targets(engine)

def bmc_remove_last_reached_targets(engine):
    """
    bmc_remove_last_reached_targets(Int_engine_bmc engine)

    Parameters
    ----------
    engine: Int_engine_bmc

    """
    return _api.bmc_remove_last_reached_targets(engine)

def mk_engine_br(ctx):
    """
    mk_engine_br(Int_ctx ctx) -> Int_engine_br

    Parameters
    ----------
    ctx: Int_ctx

    """
    return _api.mk_engine_br(ctx)

def br_add_target(ctx, engine, target):
    """
    br_add_target(Int_ctx ctx, Int_engine_br engine, Int_net target)

    Parameters
    ----------
    ctx: Int_ctx
    engine: Int_engine_br
    target: Int_net

    """
    return _api.br_add_target(ctx, engine, target)

def br_add_watch(ctx, engine, watch):
    """
    br_add_watch(Int_ctx ctx, Int_engine_br engine, Int_net watch)

    Parameters
    ----------
    ctx: Int_ctx
    engine: Int_engine_br
    watch: Int_net

    """
    return _api.br_add_watch(ctx, engine, watch)

def br_reach_targets(engine):
    """
    br_reach_targets(Int_engine_br engine) -> Int_engine_result

    Parameters
    ----------
    engine: Int_engine_br

    """
    return _api.br_reach_targets(engine)

def br_remove_last_reached_targets(engine):
    """
    br_remove_last_reached_targets(Int_engine_br engine)

    Parameters
    ----------
    engine: Int_engine_br

    """
    return _api.br_remove_last_reached_targets(engine)

def bmc_last_reached_targets_number(engine):
    """
    bmc_last_reached_targets_number(Int_engine_bmc engine) -> unsigned int

    Parameters
    ----------
    engine: Int_engine_bmc

    """
    return _api.bmc_last_reached_targets_number(engine)

def bmc_last_reached_target(engine, n):
    """
    bmc_last_reached_target(Int_engine_bmc engine, unsigned int n) -> Int_net

    Parameters
    ----------
    engine: Int_engine_bmc
    n: unsigned int

    """
    return _api.bmc_last_reached_target(engine, n)

def br_last_reached_targets_number(engine):
    """
    br_last_reached_targets_number(Int_engine_br engine) -> unsigned int

    Parameters
    ----------
    engine: Int_engine_br

    """
    return _api.br_last_reached_targets_number(engine)

def br_last_reached_target(engine, n):
    """
    br_last_reached_target(Int_engine_br engine, unsigned int n) -> Int_net

    Parameters
    ----------
    engine: Int_engine_br
    n: unsigned int

    """
    return _api.br_last_reached_target(engine, n)

def bmc_get_trace(ctx, bmc, target):
    """
    bmc_get_trace(Int_ctx ctx, Int_engine_bmc bmc, Int_net target) -> Int_trace

    Parameters
    ----------
    ctx: Int_ctx
    bmc: Int_engine_bmc
    target: Int_net

    """
    return _api.bmc_get_trace(ctx, bmc, target)

def br_get_trace(ctx, br, target):
    """
    br_get_trace(Int_ctx ctx, Int_engine_br br, Int_net target) -> Int_trace

    Parameters
    ----------
    ctx: Int_ctx
    br: Int_engine_br
    target: Int_net

    """
    return _api.br_get_trace(ctx, br, target)

def trace_prepare_value_for_net(ctx, cex, net, depth):
    """
    trace_prepare_value_for_net(Int_ctx ctx, Int_trace cex, Int_net net, unsigned int depth) -> unsigned int

    Parameters
    ----------
    ctx: Int_ctx
    cex: Int_trace
    net: Int_net
    depth: unsigned int

    """
    return _api.trace_prepare_value_for_net(ctx, cex, net, depth)

def prepare_value_for_net(ctx, net):
    """
    prepare_value_for_net(Int_ctx ctx, Int_net net) -> unsigned int

    Parameters
    ----------
    ctx: Int_ctx
    net: Int_net

    """
    return _api.prepare_value_for_net(ctx, net)

def value_at(i):
    """
    value_at(unsigned int i) -> char

    Parameters
    ----------
    i: unsigned int

    """
    return _api.value_at(i)

def trace_get_max_depth(trace):
    """
    trace_get_max_depth(Int_trace trace) -> unsigned int

    Parameters
    ----------
    trace: Int_trace

    """
    return _api.trace_get_max_depth(trace)

def mk_trace(ctx):
    """
    mk_trace(Int_ctx ctx) -> Int_trace

    Parameters
    ----------
    ctx: Int_ctx

    """
    return _api.mk_trace(ctx)

def trace_set_value(ctx, trace, net, depth, value):
    """
    trace_set_value(Int_ctx ctx, Int_trace trace, Int_net net, unsigned int depth, char const * value)

    Parameters
    ----------
    ctx: Int_ctx
    trace: Int_trace
    net: Int_net
    depth: unsigned int
    value: char const *

    """
    return _api.trace_set_value(ctx, trace, net, depth, value)

def trace_get_watched_nets_number(trace):
    """
    trace_get_watched_nets_number(Int_trace trace) -> unsigned int

    Parameters
    ----------
    trace: Int_trace

    """
    return _api.trace_get_watched_nets_number(trace)

def trace_get_watched_net(trace, i):
    """
    trace_get_watched_net(Int_trace trace, unsigned int i) -> Int_net

    Parameters
    ----------
    trace: Int_trace
    i: unsigned int

    """
    return _api.trace_get_watched_net(trace, i)

def mk_simulator(ctx):
    """
    mk_simulator(Int_ctx ctx) -> Int_simulator

    Parameters
    ----------
    ctx: Int_ctx

    """
    return _api.mk_simulator(ctx)

def simulator_add_watch(ctx, simulator, watch):
    """
    simulator_add_watch(Int_ctx ctx, Int_simulator simulator, Int_net watch)

    Parameters
    ----------
    ctx: Int_ctx
    simulator: Int_simulator
    watch: Int_net

    """
    return _api.simulator_add_watch(ctx, simulator, watch)

def simulator_simulate(simulator, trace, depth):
    """
    simulator_simulate(Int_simulator simulator, Int_trace trace, unsigned int depth)

    Parameters
    ----------
    simulator: Int_simulator
    trace: Int_trace
    depth: unsigned int

    """
    return _api.simulator_simulate(simulator, trace, depth)

def mk_boolean_type(ctx):
    """
    mk_boolean_type(Int_ctx ctx) -> Int_type

    Parameters
    ----------
    ctx: Int_ctx

    """
    return _api.mk_boolean_type(ctx)

def mk_int8_type(ctx):
    """
    mk_int8_type(Int_ctx ctx) -> Int_type

    Parameters
    ----------
    ctx: Int_ctx

    """
    return _api.mk_int8_type(ctx)

def mk_int16_type(ctx):
    """
    mk_int16_type(Int_ctx ctx) -> Int_type

    Parameters
    ----------
    ctx: Int_ctx

    """
    return _api.mk_int16_type(ctx)

def mk_int32_type(ctx):
    """
    mk_int32_type(Int_ctx ctx) -> Int_type

    Parameters
    ----------
    ctx: Int_ctx

    """
    return _api.mk_int32_type(ctx)

def mk_uint8_type(ctx):
    """
    mk_uint8_type(Int_ctx ctx) -> Int_type

    Parameters
    ----------
    ctx: Int_ctx

    """
    return _api.mk_uint8_type(ctx)

def mk_uint16_type(ctx):
    """
    mk_uint16_type(Int_ctx ctx) -> Int_type

    Parameters
    ----------
    ctx: Int_ctx

    """
    return _api.mk_uint16_type(ctx)

def mk_uint32_type(ctx):
    """
    mk_uint32_type(Int_ctx ctx) -> Int_type

    Parameters
    ----------
    ctx: Int_ctx

    """
    return _api.mk_uint32_type(ctx)

def mk_real_type(ctx):
    """
    mk_real_type(Int_ctx ctx) -> Int_type

    Parameters
    ----------
    ctx: Int_ctx

    """
    return _api.mk_real_type(ctx)

def mk_float16_type(ctx):
    """
    mk_float16_type(Int_ctx ctx) -> Int_type

    Parameters
    ----------
    ctx: Int_ctx

    """
    return _api.mk_float16_type(ctx)

def mk_float32_type(ctx):
    """
    mk_float32_type(Int_ctx ctx) -> Int_type

    Parameters
    ----------
    ctx: Int_ctx

    """
    return _api.mk_float32_type(ctx)

def mk_float64_type(ctx):
    """
    mk_float64_type(Int_ctx ctx) -> Int_type

    Parameters
    ----------
    ctx: Int_ctx

    """
    return _api.mk_float64_type(ctx)

def mk_undef(ctx):
    """
    mk_undef(Int_ctx ctx) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx

    """
    return _api.mk_undef(ctx)

def mk_true(ctx):
    """
    mk_true(Int_ctx ctx) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx

    """
    return _api.mk_true(ctx)

def mk_false(ctx):
    """
    mk_false(Int_ctx ctx) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx

    """
    return _api.mk_false(ctx)

def mk_number(ctx, value, type):
    """
    mk_number(Int_ctx ctx, char const * value, Int_type type) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx
    value: char const *
    type: Int_type

    """
    return _api.mk_number(ctx, value, type)

def mk_not(ctx, x):
    """
    mk_not(Int_ctx ctx, Int_net x) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx
    x: Int_net

    """
    return _api.mk_not(ctx, x)

def mk_and(ctx, x, y):
    """
    mk_and(Int_ctx ctx, Int_net x, Int_net y) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx
    x: Int_net
    y: Int_net

    """
    return _api.mk_and(ctx, x, y)

def mk_or(ctx, x, y):
    """
    mk_or(Int_ctx ctx, Int_net x, Int_net y) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx
    x: Int_net
    y: Int_net

    """
    return _api.mk_or(ctx, x, y)

def mk_xor(ctx, x, y):
    """
    mk_xor(Int_ctx ctx, Int_net x, Int_net y) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx
    x: Int_net
    y: Int_net

    """
    return _api.mk_xor(ctx, x, y)

def mk_ite(ctx, i, t, e):
    """
    mk_ite(Int_ctx ctx, Int_net i, Int_net t, Int_net e) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx
    i: Int_net
    t: Int_net
    e: Int_net

    """
    return _api.mk_ite(ctx, i, t, e)

def mk_iff(ctx, x, y):
    """
    mk_iff(Int_ctx ctx, Int_net x, Int_net y) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx
    x: Int_net
    y: Int_net

    """
    return _api.mk_iff(ctx, x, y)

def mk_minus(ctx, x):
    """
    mk_minus(Int_ctx ctx, Int_net x) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx
    x: Int_net

    """
    return _api.mk_minus(ctx, x)

def mk_add(ctx, x, y):
    """
    mk_add(Int_ctx ctx, Int_net x, Int_net y) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx
    x: Int_net
    y: Int_net

    """
    return _api.mk_add(ctx, x, y)

def mk_sub(ctx, x, y):
    """
    mk_sub(Int_ctx ctx, Int_net x, Int_net y) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx
    x: Int_net
    y: Int_net

    """
    return _api.mk_sub(ctx, x, y)

def mk_mul(ctx, x, y):
    """
    mk_mul(Int_ctx ctx, Int_net x, Int_net y) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx
    x: Int_net
    y: Int_net

    """
    return _api.mk_mul(ctx, x, y)

def mk_div(ctx, x, y):
    """
    mk_div(Int_ctx ctx, Int_net x, Int_net y) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx
    x: Int_net
    y: Int_net

    """
    return _api.mk_div(ctx, x, y)

def mk_mod(ctx, x, y):
    """
    mk_mod(Int_ctx ctx, Int_net x, Int_net y) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx
    x: Int_net
    y: Int_net

    """
    return _api.mk_mod(ctx, x, y)

def mk_eq(ctx, x, y):
    """
    mk_eq(Int_ctx ctx, Int_net x, Int_net y) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx
    x: Int_net
    y: Int_net

    """
    return _api.mk_eq(ctx, x, y)

def mk_leq(ctx, x, y):
    """
    mk_leq(Int_ctx ctx, Int_net x, Int_net y) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx
    x: Int_net
    y: Int_net

    """
    return _api.mk_leq(ctx, x, y)

def mk_lt(ctx, x, y):
    """
    mk_lt(Int_ctx ctx, Int_net x, Int_net y) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx
    x: Int_net
    y: Int_net

    """
    return _api.mk_lt(ctx, x, y)

def mk_geq(ctx, x, y):
    """
    mk_geq(Int_ctx ctx, Int_net x, Int_net y) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx
    x: Int_net
    y: Int_net

    """
    return _api.mk_geq(ctx, x, y)

def mk_gt(ctx, x, y):
    """
    mk_gt(Int_ctx ctx, Int_net x, Int_net y) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx
    x: Int_net
    y: Int_net

    """
    return _api.mk_gt(ctx, x, y)

def mk_neq(ctx, x, y):
    """
    mk_neq(Int_ctx ctx, Int_net x, Int_net y) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx
    x: Int_net
    y: Int_net

    """
    return _api.mk_neq(ctx, x, y)

def mk_input(ctx, name, type):
    """
    mk_input(Int_ctx ctx, char const * name, Int_type type) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx
    name: char const *
    type: Int_type

    """
    return _api.mk_input(ctx, name, type)

def mk_output(arg1, net):
    """
    mk_output(Int_ctx arg1, Int_net net)

    Parameters
    ----------
    arg1: Int_ctx
    net: Int_net

    """
    return _api.mk_output(arg1, net)

def mk_assumption(arg1, net):
    """
    mk_assumption(Int_ctx arg1, Int_net net)

    Parameters
    ----------
    arg1: Int_ctx
    net: Int_net

    """
    return _api.mk_assumption(arg1, net)

def mk_latch(ctx, name, type):
    """
    mk_latch(Int_ctx ctx, char const * name, Int_type type) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx
    name: char const *
    type: Int_type

    """
    return _api.mk_latch(ctx, name, type)

def set_latch_init_next(ctx, latch, init, next):
    """
    set_latch_init_next(Int_ctx ctx, Int_net latch, Int_net init, Int_net next)

    Parameters
    ----------
    ctx: Int_ctx
    latch: Int_net
    init: Int_net
    next: Int_net

    """
    return _api.set_latch_init_next(ctx, latch, init, next)

def mk_substitute(ctx, term, new_subterm, old_subterm):
    """
    mk_substitute(Int_ctx ctx, Int_net term, Int_net new_subterm, Int_net old_subterm) -> Int_net

    Parameters
    ----------
    ctx: Int_ctx
    term: Int_net
    new_subterm: Int_net
    old_subterm: Int_net

    """
    return _api.mk_substitute(ctx, term, new_subterm, old_subterm)

def apitrace_dump_to_file(filename):
    """
    apitrace_dump_to_file(char const * filename)

    Parameters
    ----------
    filename: char const *

    """
    return _api.apitrace_dump_to_file(filename)

def apitrace_print_to_stdout():
    """apitrace_print_to_stdout()"""
    return _api.apitrace_print_to_stdout()

def apitrace_print_to_stderr():
    """apitrace_print_to_stderr()"""
    return _api.apitrace_print_to_stderr()
# This file is compatible with both classic and new-style classes.





"""
Copyright (C) 2017 Roberto Bruttomesso <roberto.bruttomesso@gmail.com>

This file is distributed under the terms of the 3-clause BSD License.
A copy of the license can be found in the root directory or at
https://opensource.org/licenses/BSD-3-Clause.

Author: Roberto Bruttomesso <roberto.bruttomesso@gmail.com>
  Date: 27/03/2017

This module implements infrastructure to store instructions
"""

import intrepyd as ip
from intrepyd.engine import EngineResult
import pandas as pd
import collections as cl

def compute_mcdc(context, class_, decisions, maxDepth):
    """
    Computes MC/DC tests in form of a table.

    Args:
        context: the intrepyd context to use
        class_ (class): a python class that defines the circuit
        decisions (Dictionary): each key of the dictionary is the
                    name of a decision, and the values are the
                    corresponding conditions

                { netName           : [netName, ...] }
                  ^                    ^
                  the decision         the list of conditions

    Returns:
        (Dictionary) the MC/DC tables, one per each decision
    """

    # Fetches and duplicates the circuit
    instA = class_(context, 'InstA')
    instB = class_(context, 'InstB')
    instA.mk_circuit(True)
    instB.mk_circuit(True)

    # Flattens inputs and outputs into nets, for simplicity
    instA.nets.update(instA.inputs)
    instB.nets.update(instB.inputs)

    # Creates test objectives
    decision2testObjectives = { decision :\
            compute_mcdc_targets(context, instA, instB, decision, conditions)\
            for decision, conditions in decisions.iteritems() }

    # Compute MC/DC traces: this is the computationally
    # expensive part that calls model checking routines
    decision2traces, decision2unreachable = solve_mcdc_targets(context, decision2testObjectives, maxDepth)

    # Compute MC/DC tables from traces
    decision2table = compute_mcdc_tables(context, instA, instB, decision2traces, decisions)

    # Compute pretty tables
    decision2prettyTable = compute_pretty_tables(decision2table)

    return decision2prettyTable, decision2unreachable

def compute_mcdc_targets(context, instA, instB, decision, conditions):
    """
    Computes the MC/DC reachability target for the
    two copies of the circuit.

    Args:
        context: the intrepyd context to use
        instA: an instance of the circuit
        instB: an instance of the circuit
        decision: the name of a decision
        conditions: the list of names of the conditions

    Returns:
        the list of targets, whose reachability implies the existence of an MC/DC test
    """
    decisionA = instA.nets[decision]
    decisionB = instB.nets[decision]
    decisionA_diff_decisionB = context.mk_neq(decisionA, decisionB)

    conditionA_neq_conditionB = []
    conditionsA = []
    conditionsB = []
    for condition in conditions:
        cA = instA.nets[condition]
        conditionsA.append(cA)
        cB = instB.nets[condition]
        conditionsB.append(cB)
        conditionA_neq_conditionB.append(context.mk_neq(cA, cB))

    targets = []
    for i in range(len(conditions)):
        # Building test objective for the i-th condition
        cA = conditionsA[i]
        cB = conditionsB[i]
        # i-th condition must differ in A and B
        conj1 = conditionA_neq_conditionB[i]
        # test objective must differ in A and B
        conj2 = decisionA_diff_decisionB
        # decisionA[not(cA)/cA] != decisionA
        not_cA = context.mk_not(cA)
        decisionA_not_cA = context.mk_substitute(decisionA, not_cA, cA)
        conj3 = context.mk_neq(decisionA_not_cA, decisionA)
        # decisionB[not(cB)/cB] != decisionB
        not_cB = context.mk_not(cB)
        decisionB_not_cB = context.mk_substitute(decisionB, not_cB, cB)
        conj4 = context.mk_neq(decisionB_not_cB, decisionB)
        # Creates final conjunction
        tmp1 = context.mk_and(conj1, conj2)
        tmp2 = context.mk_and(tmp1, conj3)
        target = context.mk_and(tmp2, conj4)
        targets.append(target)

    return targets

def solve_mcdc_targets(context, decision2testObjectives, maxDepth):
    """
    Solves the MC/DC targets, maximizing the number of solved
    test objectives per each call.
    """
    # Bmc engine will be used to compute counterexamples
    bmc = context.mk_optimizing_bmc()

    testObjectives2decision = {}
    decision2traces = {}
    decision2unreachable = {}
    targets = []
    for decision, tos in decision2testObjectives.iteritems():
        decision2traces[decision] = []
        decision2unreachable[decision] = []
        for to in tos:
            testObjectives2decision[to] = decision
            targets.append(to)

    totalTargets = len(targets)

    # Compute unreachable targets first
    totalUnreached = 0
    totalReached = 0
    for target in targets:
        br = context.mk_backward_reach()
        br.add_target(target)
        result = br.reach_targets()
        if result == EngineResult.UNREACHABLE:
            decision = testObjectives2decision[target]
            decision2unreachable[decision].append(target)
            totalUnreached += 1
        elif result == EngineResult.REACHABLE:
            bmc.add_target(target)
            totalReached += 1

    # print 'There are', totalTargets, 'test objectives:'
    # print '-', totalUnreached, 'unreachable test objectives'
    # print '-', totalReached, 'reachable test objectives'

    if totalUnreached == totalTargets:
        return decision2traces, decision2unreachable

    # Compute counterexamples for reachable targets
    done = False
    depth = 0
    bmcTotalReached = 0
    while not done:
        bmc.set_current_depth(depth)
        result = bmc.reach_targets()
        if result != EngineResult.REACHABLE:
            if depth == maxDepth:
                done = True
            depth += 1
            continue
        reached = bmc.get_last_reached_targets()
        trace = bmc.get_last_trace()
        for reachedTarget in reached:
            decision = testObjectives2decision[reachedTarget]
            decision2traces[decision].append(trace)
            bmcTotalReached += 1
        bmc.remove_last_reached_targets()
        if bmcTotalReached == totalReached:
            done = True

    totalUndecided = totalReached - bmcTotalReached
    if totalUndecided != 0:
        print 'There are', totalUndecided, 'undecided test objectives within depth', maxDepth

    return decision2traces, decision2unreachable

def compute_mcdc_tables(context, instA, instB, decision2traces, decisions):
    """
    Computes MC/DC tables out of counterexamples.
    """
    decision2table = {}
    for decision, traces in decision2traces.iteritems():
        instAtestNets = instA.inputs.values() + [instA.nets[decision]]
        instBtestNets = instB.inputs.values() + [instB.nets[decision]]
        header = [cond for cond in decisions[decision]]
        header.append(decision)
        decision2table[decision] = [header]
        for trace in traces:
            simulator = context.mk_simulator()
            simulator.add_watch(instA.nets[decision])
            simulator.add_watch(instB.nets[decision])
            simulator.simulate(trace, trace.get_max_depth())
            fullTestA = trace.get_as_net_dictionary()
            fullTestB = trace.get_as_net_dictionary()
            testA = [v[0] for k, v in fullTestA.iteritems() if k in instAtestNets]
            testB = [v[0] for k, v in fullTestB.iteritems() if k in instBtestNets]
            decision2table[decision].append((testA, testB))
    return decision2table

def compute_pretty_tables(decision2table):
    """
    Postprocess raw MC/DC tables to get something presentation ready.
    """
    decision2prettyTable = {}
    first = True
    for decision, table in decision2table.iteritems():
        prettyTable = []
        for row in table:
            if first:
                first = False
                prettyTable.append(row)
            else:
                prettyTable.append(row[0])
                prettyTable.append(row[1])
        decision2prettyTable[decision] = prettyTable
    return decision2prettyTable

def get_tables_as_dataframe(decision2table):
    result = {}
    for decision, table in decision2table.iteritems():
        if len(table) == 1:
            continue
        df = pd.DataFrame(table[1:], columns=table[0])
        df = df.drop_duplicates()
        result[decision] = df
    return result



"""
Copyright (C) 2017 Roberto Bruttomesso <roberto.bruttomesso@gmail.com>

This file is distributed under the terms of the 3-clause BSD License.
A copy of the license can be found in the root directory or at
https://opensource.org/licenses/BSD-3-Clause.

Author: Roberto Bruttomesso <roberto.bruttomesso@gmail.com>
  Date: 27/03/2017
"""
import unittest
import intrepyd
import intrepyd.circuit
from intrepyd.engine import EngineResult

class MyCirc(intrepyd.circuit.Circuit):
    
    def __init__(self, context, name):
        intrepyd.circuit.Circuit.__init__(self, context, name)
   
    def _mk_naked_circuit_impl(self, inputs):
        inputKeys = list(inputs)
        i1 = inputs[inputKeys[0]]
        i2 = inputs[inputKeys[1]]
        output = self.context.mk_and(i1, i2)
        return output
        

class TestCircuit(unittest.TestCase):

    def test_circuit_01(self):
        ctx = intrepyd.Context()
        inst1 = MyCirc(ctx, 'mycirc1')
        inst2 = MyCirc(ctx, 'mycirc2')
        bt = ctx.mk_boolean_type()
        i1 = ctx.mk_input('i1', bt)
        i2 = ctx.mk_input('i2', bt)
        inputs = { 'i1' : i1, 'i2' : i2 }
        out1 = inst1.mk_naked_circuit(inputs, True) 
        out2 = inst2.mk_naked_circuit(inputs, True) 
        diff = ctx.mk_xor(out1, out2)
        bmc = ctx.mk_bmc()
        bmc.add_target(diff)
        result = bmc.reach_targets()   
        self.assertEqual(EngineResult.UNREACHABLE, result)

if __name__ == '__main__':
    unittest.main()



"""
Copyright (C) 2017 Roberto Bruttomesso <roberto.bruttomesso@gmail.com>

This file is distributed under the terms of the 3-clause BSD License.
A copy of the license can be found in the root directory or at
https://opensource.org/licenses/BSD-3-Clause.

Author: Roberto Bruttomesso <roberto.bruttomesso@gmail.com>
  Date: 27/03/2017

This module implements some color shortcuts
"""
from colorama import Fore, Style

def fail(message):
    """
    Color message as failure
    """
    return _fmt(Fore.RED, Style.BRIGHT, message)

def warn(message):
    """
    Color message as warning
    """
    return _fmt(Fore.YELLOW, Style.BRIGHT, message)

def good(message):
    """
    Color message as success
    """
    return _fmt(Fore.GREEN, Style.BRIGHT, message)

def info(message):
    """
    Color message as info
    """
    return _fmt(Fore.CYAN, Style.BRIGHT, message)

def dim(message):
    """
    dim message
    """
    return _fmt(Fore.BLACK, Style.BRIGHT, message)

def enhance(message):
    """
    enhance message
    """
    return _fmt(Fore.WHITE, Style.BRIGHT, message)

def _fmt(color, style, message):
    """
    Helper routine
    """
    return color + style + message + Fore.RESET + Style.RESET_ALL



"""
Copyright (C) 2017 Roberto Bruttomesso <roberto.bruttomesso@gmail.com>

This file is distributed under the terms of the 3-clause BSD License.
A copy of the license can be found in the root directory or at
https://opensource.org/licenses/BSD-3-Clause.

Author: Roberto Bruttomesso <roberto.bruttomesso@gmail.com>
  Date: 27/03/2017

This module implements infrastructure to store instructions
"""

import intrepyd as ip
import intrepyd.api

def mk_counter(context, name, type, limit, init=None, increment=None, enable=None, reset=None):
    """
    Counts from init to limit by increment. When limit
    is reached, limit is outputted. An enable signal may
    also be specified.

    Args:
        context: the context to use
        name: the unique name of this counter
        type: the type of the data stored in the counter
        limit: the limit after which the counter saturates
        init (=0): the initial value of the counter
        increment (=1): the increment value
        enable (=true): the counter is incremented only if enable is true
        reset (=false): resets the counter to the initial value

    Returns:
        the current value of the counter
        true iff the counter has reached saturation
    """
    if init == None:
        init = context.mk_number("0", type)
    if increment == None:
        increment = context.mk_number("1", type)
    if enable == None:
        enable = context.mk_true()
    if reset == None:
        reset = context.mk_false()

    counter = context.mk_latch(name, type)
    notQ = context.mk_lt(counter, limit)
    next = context.mk_ite(reset,\
                          init,\
                          context.mk_ite(context.mk_and(enable, notQ),\
                                         context.mk_add(counter, increment),\
                                         counter))
    context.set_latch_init_next(counter, init, next)
    Q = context.mk_not(notQ, name + '.Q')
    return counter, Q


"""
Copyright (C) 2017 Roberto Bruttomesso <roberto.bruttomesso@gmail.com>

This file is distributed under the terms of the 3-clause BSD License.
A copy of the license can be found in the root directory or at
https://opensource.org/licenses/BSD-3-Clause.

Author: Roberto Bruttomesso <roberto.bruttomesso@gmail.com>
  Date: 27/03/2017

This module implements infrastructure for creating nets and engines
"""

import intrepyd as ip
import intrepyd.api
import intrepyd.engine
import intrepyd.simulator
import intrepyd.trace
import collections

class Context(object):
    """
    An intrepyd context
    """
    def __init__(self):
        self.ctx = ip.api.mk_ctx()
        self.inputs = collections.OrderedDict()
        self.outputs = collections.OrderedDict()
        self.latches = {}
        self.nets = {}
        self.net2name = {}
        self.input2type = {}
        self.booleantype = ip.api.mk_boolean_type(self.ctx)
        self.int8type = ip.api.mk_int8_type(self.ctx)
        self.int16type = ip.api.mk_int16_type(self.ctx)
        self.int32type = ip.api.mk_int32_type(self.ctx)
        self.uint8type = ip.api.mk_uint8_type(self.ctx)
        self.uint16type = ip.api.mk_uint16_type(self.ctx)
        self.uint32type = ip.api.mk_uint32_type(self.ctx)
        self.realtype = ip.api.mk_real_type(self.ctx)
        self.float16type = ip.api.mk_float16_type(self.ctx)
        self.float32type = ip.api.mk_float32_type(self.ctx)
        self.float64type = ip.api.mk_float64_type(self.ctx)
        self.undef = ip.api.mk_undef(self.ctx)
        self.true = ip.api.mk_true(self.ctx)
        self.false = ip.api.mk_false(self.ctx)
        self.namespaces = []

    def __del__(self):
        ip.api.del_ctx(self.ctx)

    def push_namespace(self, name):
        ip.api.push_namespace(self.ctx, name)
        self.namespaces.append(name)

    def pop_namespace(self):
        ip.api.pop_namespace(self.ctx)
        if len(self.namespaces) == 0:
            raise Exception('Cannot pop namespace, empty list')
        return self.namespaces.pop()

    def mk_boolean_type(self):
       return self.booleantype

    def mk_int8_type(self):
        return self.int8type

    def mk_int16_type(self):
        return self.int16type

    def mk_int32_type(self):
        return self.int32type

    def mk_uint8_type(self):
        return self.uint8type

    def mk_uint16_type(self):
        return self.uint16type

    def mk_uint32_type(self):
        return self.uint32type

    def mk_real_type(self):
        return self.realtype

    def mk_float16_type(self):
        return self.float16type

    def mk_float32_type(self):
        return self.float32type

    def mk_float64_type(self):
        return self.float64type

    def mk_undef(self):
        return self.undef

    def mk_true(self, name=None):
        return self._register(self.true, name)

    def mk_false(self, name=None):
        return self._register(self.false, name)

    def mk_number(self, value, type_, name=None):
        return self._register(ip.api.mk_number(self.ctx, value, type_), name)

    def mk_not(self, x, name=None):
        return self._register(ip.api.mk_not(self.ctx, x), name=name)

    def mk_and(self, x, y, name=None):
        return self._register(ip.api.mk_and(self.ctx, x, y), name=name)

    def mk_or(self, x, y, name=None):
        return self._register(ip.api.mk_or(self.ctx, x, y), name=name)

    def mk_xor(self, x, y, name=None):
        return self._register(ip.api.mk_xor(self.ctx, x, y), name=name)

    def mk_implies(self, x, y, name=None):
        return self._register(ip.api.mk_or(self.ctx, ip.api.mk_not(self.ctx, x), y), name=name)

    def mk_iff(self, x, y, name=None):
        return self._register(ip.api.mk_iff(self.ctx, x, y), name=name)

    def mk_eq(self, x, y, name=None):
        return self._register(ip.api.mk_eq(self.ctx, x, y), name=name)

    def mk_leq(self, x, y, name=None):
        return self._register(ip.api.mk_leq(self.ctx, x, y), name=name)

    def mk_lt(self, x, y, name=None):
        return self._register(ip.api.mk_lt(self.ctx, x, y), name=name)

    def mk_geq(self, x, y, name=None):
        return self._register(ip.api.mk_geq(self.ctx, x, y), name=name)

    def mk_gt(self, x, y, name=None):
        return self._register(ip.api.mk_gt(self.ctx, x, y), name=name)

    def mk_neq(self, x, y, name=None):
        return self._register(ip.api.mk_neq(self.ctx, x, y), name=name)

    def mk_add(self, x, y, name=None):
        return self._register(ip.api.mk_add(self.ctx, x, y), name=name)

    def mk_mul(self, x, y, name=None):
        return self._register(ip.api.mk_mul(self.ctx, x, y), name=name)

    def mk_div(self, x, y, name=None):
        return self._register(ip.api.mk_div(self.ctx, x, y), name=name)

    def mk_mod(self, x, y, name=None):
        return self._register(ip.api.mk_mod(self.ctx, x, y), name=name)

    def mk_sub(self, x, y, name=None):
        return self._register(ip.api.mk_sub(self.ctx, x, y), name=name)

    def mk_minus(self, x, name=None):
        return self._register(ip.api.mk_minus(self.ctx, x), name=name)

    def mk_ite(self, i, t, e, name=None):
        return self._register(ip.api.mk_ite(self.ctx, i, t, e), name=name)

    def mk_input(self, name, type_):
        return self._register_input(ip.api.mk_input(self.ctx, name, type_), type_, name=name)

    def mk_output(self, x, name=None):
        ip.api.mk_output(self.ctx, x)
        self._register_output(x, name=name)

    def mk_latch(self, name, type_):
        return self._register_latch(ip.api.mk_latch(self.ctx, name, type_), name=name)

    def set_latch_init_next(self, latch, init, next):
        ip.api.set_latch_init_next(self.ctx, latch, init, next)

    def mk_substitute(self, term, newTerm, oldTerm):
        return ip.api.mk_substitute(self.ctx, term, newTerm, oldTerm)

    def mk_assumption(self, net):
        ip.api.mk_assumption(self.ctx, net)

    def mk_bmc(self):
        return ip.engine.Bmc(self.ctx)

    def mk_optimizing_bmc(self):
        return ip.engine.OptimizingBmc(self.ctx)

    def mk_backward_reach(self):
        return ip.engine.BackwardReach(self.ctx)

    def mk_simulator(self):
        return ip.simulator.Simulator(self.ctx)

    def mk_trace(self):
        return ip.trace.Trace(self.ctx)

    def to_string(self, net):
        """
        Returns the given net as a string, as given from the underlying smt-solver.
        """
        size = ip.api.prepare_value_for_net(self.ctx, net)
        value = ''
        for i in range(size):
            value += ip.api.value_at(i)
        return value

    def get_default_value(self, type_):
        """
        Returns a default value for the given type
        """
        if type_ == self.booleantype:
            return 'F'
        elif type_ in [self.float16type, self.float32type, self.float64type, self.realtype]:
            return '0.0'
        assert type_ in [self.int8type, self.int16type, self.int32type, self.uint8type, self.uint16type, self.uint32type]
        return '0'

    def _current_namespace_prefix(self):
        result = ''
        for namespace in self.namespaces:
            result += namespace + '.'
        return result

    def _register(self, rawnet, name):
        if rawnet in self.net2name:
            return rawnet
        if name is None:
            name = '__n' + str(rawnet)
        name = self._current_namespace_prefix() + name
        if name in self.nets:
            raise Exception('Name already used: ' + name)
        self.nets[name] = rawnet
        self.net2name[rawnet] = name
        return rawnet

    def _register_input(self, rawnet, type_, name):
        rawnet = self._register(rawnet, name)
        self.inputs[name] = rawnet
        self.input2type[rawnet] = type_
        return rawnet

    def _register_latch(self, rawnet, name):
        rawnet = self._register(rawnet, name)
        self.latches[name] = rawnet
        return rawnet

    def _register_output(self, rawnet, name):
        if name is None:
            name = '__o' + str(rawnet)
        rawnet = self._register(rawnet, name)
        self.outputs[name] = rawnet



"""
Copyright (C) 2017 Roberto Bruttomesso <roberto.bruttomesso@gmail.com>

This file is distributed under the terms of the 3-clause BSD License.
A copy of the license can be found in the root directory or at
https://opensource.org/licenses/BSD-3-Clause.

Author: Roberto Bruttomesso <roberto.bruttomesso@gmail.com>
  Date: 27/03/2017

This module implements infrastructure to store nodes
"""

from intrepyd.lustre2py.visitable import Visitable
from intrepyd.lustre2py.expression import Expression

_GREY = 1
_BLACK = 2

class Node(Visitable):
    """
    Stores a node
    """
    def __init__(self, name):
        self._name = name
        self._input_decls = []
        self._output_decls = []
        self._local_decls = []
        self._equations = []
        self._main = False
        self._properties = []

    @property
    def name(self):
        """
        Getter
        """
        return self._name

    @property
    def input_decls(self):
        """
        Getter
        """
        return self._input_decls

    @property
    def output_decls(self):
        """
        Getter
        """
        return self._output_decls

    @property
    def local_decls(self):
        """
        Getter
        """
        return self._local_decls

    @property
    def equations(self):
        """
        Getter
        """
        return self._equations

    @property
    def main(self):
        """
        Getter
        """
        return self._main

    @property
    def properties(self):
        """
        Getter
        """
        return self._properties

    @name.setter
    def name(self, value):
        self._name = value

    @main.setter
    def main(self, value):
        self._main = value

    def add_input_decl(self, inputs):
        """
        Add a new input declaration
        """
        self._input_decls.append(inputs)

    def add_output_decl(self, outputs):
        """
        Add a new output declaration
        """
        self._output_decls.append(outputs)

    def add_local_decl(self, llocals):
        """
        Add a new local declaration
        """
        self._local_decls.append(llocals)

    def add_equation(self, equation):
        """
        Adds an equation
        """
        self._equations.append(equation)

    def add_property(self, prop):
        """
        Adds a property declaration
        """
        self._properties.append(prop)

    def sort_equations(self):
        """
        Sort equations in such a way that local variables are
        defined before their use.
        """
        eqtopology = EquationTopology(self.equations, self.local_decls + self.output_decls)
        topological_order = eqtopology.compute_topological_order()
        new_equations = [self.equations[i] for i in topological_order]
        self._equations = new_equations

    def push_pre_in_equations(self):
        """
        Rewrites rhs of equations in such a way that 'pre' operators
        are pushed to variables
        """
        for equat in self._equations:
            equat.push_pre()


class EquationTopology(object):
    """
    Utility class for sorting equation dependencies topologically
    """
    def __init__(self, equations, local_decls):
        self._equations = equations
        self._local_decls = local_decls

    def compute_topological_order(self):
        """
        Compute dependencies between equations w.r.t. use of local variables

        0. a = b + c
        1. b = c
        2. c = 5 + d

        0 -> [2, 3]
        1 -> [2]
        2 -> []       (d is not a local variable)
        """
        dependencies = self._compute_equation_dependencies()
        return self._compute_topological_order(dependencies)

    def _compute_topological_order(self, dependencies):
        seen = {}
        topological_order = []
        for eqnum, _ in dependencies.iteritems():
            parent = {eqnum : None}
            self._compute_topological_order_rec(eqnum, dependencies,\
                                                topological_order, seen, parent)
        return topological_order

    def _compute_topological_order_rec(self, eqnum, dependencies,\
                                       topological_order, seen, parent):
        if eqnum in seen:
            if seen[eqnum] == _GREY:
                print 'Dependency loop:', parent
                raise Exception('Loop in equation dependencies')
            elif seen[eqnum] == _BLACK:
                # Already seen
                return
        seen[eqnum] = _GREY
        # Visit dependencies first
        for dependency in dependencies[eqnum]:
            if eqnum == dependency:
                continue
            parent[dependency] = eqnum
            self._compute_topological_order_rec(dependency, dependencies,\
                                                topological_order, seen, parent)
        topological_order.append(eqnum)
        seen[eqnum] = _BLACK

    def _compute_equation_dependencies(self):
        localvars = self._retrieve_local_vars()
        var2equation = self._compute_var2equation(localvars)
        dependencies = {}
        index = 0
        for equat in self._equations:
            dependencies[index] = set(self._retrieve_used_equations(equat.expression, var2equation))
            index += 1
        return dependencies

    def _retrieve_used_equations(self, expr, var2equation):
        """
        Computes the equations that expr depends upon. For
        each variable in expr, store the equation where
        the variable is defined
        """
        used_equations = []
        if expr.kind == Expression.ZEROARY:
            if expr.name in var2equation:
                used_equations.append(var2equation[expr.name])
        elif expr.kind == Expression.UNARY:
            if expr.operator != 'pre':
                used_equations = self._retrieve_used_equations(expr.operand, var2equation)
        elif expr.kind == Expression.BINARY:
            used_equations = self._retrieve_used_equations(expr.left, var2equation) +\
                             self._retrieve_used_equations(expr.right, var2equation)
        elif expr.kind == Expression.ITE:
            used_equations = self._retrieve_used_equations(expr.if_, var2equation) +\
                             self._retrieve_used_equations(expr.then_, var2equation) +\
                             self._retrieve_used_equations(expr.else_, var2equation)
        elif expr.kind == Expression.INITCURR:
            used_equations = self._retrieve_used_equations(expr.init, var2equation) +\
                             self._retrieve_used_equations(expr.curr, var2equation)
        elif expr.kind == Expression.CALL:
            for param in expr.params:
                used_equations += self._retrieve_used_equations(param, var2equation)
        elif expr.kind == Expression.TUPLE:
            for expression in expr.expressions:
                used_equations += self._retrieve_used_equations(expression, var2equation)
        elif expr.kind == Expression.LITERAL:
            pass
        else:
            raise TypeError('Unhandled kind' + expr.kind)
        return used_equations

    def _retrieve_local_vars(self):
        local_vars = []
        for decl in self._local_decls:
            for var in decl.variables:
                local_vars.append(var)
        return set(local_vars)

    def _compute_var2equation(self, localvars):
        """
        Computes a dictionary that maps a variable
        on the lhs of an equation with the number of the
        equation itself
        """
        var2equation = {}
        index = 0
        for equat in self._equations:
            for lhs in equat.lhs:
                if lhs in localvars:
                    var2equation[lhs] = index
            index += 1
        return var2equation



"""
Copyright (C) 2017 Roberto Bruttomesso <roberto.bruttomesso@gmail.com>

This file is distributed under the terms of the 3-clause BSD License.
A copy of the license can be found in the root directory or at
https://opensource.org/licenses/BSD-3-Clause.

Author: Roberto Bruttomesso <roberto.bruttomesso@gmail.com>
  Date: 27/03/2017

This module implements infrastructure to store instructions
"""

import intrepyd as ip
import collections

class Circuit(object):
    """
    Abstract interface for a circuit.
    """

    def __init__(self, context, name):
        self.context = context
        self.name = name
        self.nets = collections.OrderedDict()
        self.inputs = collections.OrderedDict()
        self.outputs = collections.OrderedDict()
        self.targets = collections.OrderedDict()

    def mk_circuit(self, putNamespace=False):
        """
        Makes the circuit, including inputs and outputs.
        Optionally, it inserts a namespace.

        Args:
            namespace (bool=False): whether to set a namespace
        """
        if putNamespace:
            self.context.push_namespace(self.name)
        self._mk_inputs()
        self.outputs = self._mk_naked_circuit_impl(self.inputs)
        self._mk_outputs()
        if putNamespace:
            self.context.pop_namespace()

    def mk_naked_circuit(self, inputs, namespace=False):
        """
        Makes the naked circuit, using the provided inputs.
        It fills out the dictionaries with the created nets.
        Optionally, it inserts a namespace.

        Args:
            inputs (OrderedDict): the inputs to use
            namespace (bool=False): whether to set a namespace

        Returns:
            (OrderedDict) the ordered dictionary of the outputs
        """
        if namespace:
            self.context.push_namespace(self.name)
        outputs = self._mk_naked_circuit_impl(inputs)
        if namespace:
            self.context.pop_namespace()
        return outputs

    def _mk_outputs(self):
        """
        Makes the outputs contained in the already filled-out
        self.outputs.
        """
        for i, (name, net) in enumerate(self.outputs.iteritems()):
            self.context.mk_output(net)

    def _mk_inputs(self):
        """
        Makes the inputs nets, and it fills the self.inputs 
        dictionary with the inputs nets.
        """
        raise NotImplementedError('Should have implemented this')

    def _mk_naked_circuit_impl(self, inputs):
        """
        Makes the naked circuit.

        Args:
            inputs (OrderedDict): the inputs to use

        Returns:
            (OrderedDict) the outputs
        """
        raise NotImplementedError('Should have implemented this')



"""
Copyright (C) 2017 Roberto Bruttomesso <roberto.bruttomesso@gmail.com>

This file is distributed under the terms of the 3-clause BSD License.
A copy of the license can be found in the root directory or at
https://opensource.org/licenses/BSD-3-Clause.

Author: Roberto Bruttomesso <roberto.bruttomesso@gmail.com>
  Date: 07/05/2017

Translates a lustre file into an intrepyd equivalent
"""

import datetime
import intrepyd.lustre2py.parser as lusp
from intrepyd.lustre2py.ast2intrepyd import Ast2Intrepyd, TAB, CONTEXT, FIRSTTICK,\
                                            LUSTREDT2INTREPYDDT, BOOLTYPE, INTTYPE,\
                                            REALTYPE, LATCH2PRE, LATCHEQUIV

def compute_node_prototype(node):
    """
    Extracts node prototype
    """
    inputs = []
    for decl in node.input_decls:
        inputs += [decl.datatype.name] * len(decl.variables)
    outputs = []
    for decl in node.output_decls:
        outputs += [decl.datatype.name] * len(decl.variables)
    return (inputs, outputs)

def translate(filename, topnode, outfilename, realtype):
    """
    Translates a lustre file into a python
    """
    if not realtype in ['real', 'float16', 'float32', 'float64']:
        raise Exception('Unsupported real type ' + realtype)
    # Parse lustre into AST
    ast = lusp.parse(filename)
    # Compute nodes prototypes
    node2proto = {node.name : compute_node_prototype(node) for node in ast.nodes}
    # Translate AST into intrepyd
    encoding = ''
    properties = []
    top = None
    ast_printer = Ast2Intrepyd(node2proto, realtype)
    for node in ast.nodes:
        # Important: sort equations by dependencies!
        node.sort_equations()
        # Important: push pre operators to leaves!
        node.push_pre_in_equations()
        # ast_printer = AstPrinter()
        # print node.accept(ast_printer)
        result, props = node.accept(ast_printer)
        encoding = result
        properties += props
        if node.main:
            top = node
        elif node.name == topnode:
            top = node
    if top is None:
        raise Exception('Top node not found')

    with open(outfilename, 'w') as outfile:
        today = str(datetime.date.today())
        outfile.write('# Translated from ' + filename + ' using intrepyd.lustre2py on ' + today)
        outfile.write('\n\n')
        outfile.write('import intrepyd\n\n')
        outfile.write(LATCH2PRE + ' = {}\n')
        outfile.write(LATCHEQUIV + ' = []\n\n')
        outfile.write(encoding)
        outfile.write('def lustre2py_main(' + CONTEXT +'):\n')
        outfile.write(TAB + BOOLTYPE + ' = ' + CONTEXT + '.mk_boolean_type()\n')
        outfile.write(TAB + INTTYPE + ' = ' + CONTEXT + '.mk_int32_type()\n')
        outfile.write(TAB + REALTYPE + ' = ' + CONTEXT + '.mk_' + realtype + '_type()\n')
        outfile.write(TAB + FIRSTTICK + ' = ' + CONTEXT +\
                      '.mk_latch("' + FIRSTTICK + '", ctx.mk_boolean_type())\n')
        outfile.write(TAB + CONTEXT +\
                      '.set_latch_init_next(' + FIRSTTICK + ', ctx.mk_true(), ctx.mk_false())\n')
        index = 0
        inputs = []
        for ttype in node2proto[top.name][0]:
            name = 'i%d' % index
            outfile.write(TAB + name + ' = ' + CONTEXT + '.mk_input("' + name + '", ' +\
                          LUSTREDT2INTREPYDDT[ttype] + ')\n')
            inputs.append(name)
            index += 1
        args = CONTEXT + ', ' + FIRSTTICK
        for inp in inputs:
            args += ', ' + inp
        index = 0
        sep = ''
        outs = ''
        if len(node2proto[top.name][1]) == 0:
            raise RuntimeError('Top node has no outputs')
        for _ in node2proto[top.name][1]:
            name = 'o%d' % index
            index += 1
            outs += sep + name
            sep = ', '
        outfile.write(TAB + outs + ' = ' + top.name + '(' + args + ')\n')
        outfile.write(TAB + 'return [' + outs + ']\n')
        outfile.write('\ndef lustre2py_main_ctxless():\n')
        outfile.write(TAB + CONTEXT + ' = intrepyd.Context()\n')
        outfile.write(TAB + 'return ' + CONTEXT + ', lustre2py_main(' + CONTEXT + ')\n')
        outfile.write('\nif __name__ == "__main__":\n')
        outfile.write(TAB + 'lustre2py_main_ctxless()\n\n')
